<?php
/**
 * EthnaTask_DB.php
 *
 * @author TSURUOKA Naoya <tsuruoka@labs.cybozu.co.jp>
 */

require_once 'Ethna/class/DB/Ethna_DB_ADOdb.php';

/**
 * EthnaTask_DB
 *
 * @access public
 * @author TSURUOKA Naoya <tsuruoka@labs.cybozu.co.jp>
 */
class EthnaTask_DB extends Ethna_DB_ADOdb
{
    function getRecentTaskFromUser($user, $show_finished = false)
    {
        if ($show_finished) {
            $sql = "SELECT * FROM task WHERE user = ?";
            $sql.= " ORDER BY task_order LIMIT 30";
        } else {
            $sql = "SELECT * FROM task WHERE user = ? AND deleted_at = 0";
            $sql.= " ORDER BY task_order LIMIT 30";
        }

        return $this->getAll($sql, array($user));
    }

    function completeTaskFromId($id, $time = false)
    {
        $id = (int)$id;
        if ($time === false) {
            $time = time();
        }
        $sql = "UPDATE task SET deleted_at = ? WHERE id = ?";
        return $this->execute($sql, array($time, $id));
    }

    /**
     * deleteTaskFromId
     *
     * @param int $id
     * @todo implement delete only own task
     */
    function deleteTaskFromId($id)
    {
        $sql = "DELETE FROM task WHERE id = ?";
        return $this->getAll($sql, array((int)$id));
    }

    function getUncompleteTask()
    {
        $sql = "SELECT * FROM task WHERE deleted_at = 0";
        return $this->getAll($sql);
    }

    function addTask($param)
    {
        $time = time();
        if (!isset($param['created_at'])) {
            $param['created_at'] = $time;
        }
        if (!isset($param['updated_at'])) {
            $param['updated_at'] = $time;
        }
        $param['task_order'] = $this->getMaxOrder() + 1;

        return $this->autoExecute('task', $param, 'INSERT');
    }

    function getMaxOrder()
    {
        $sql = "SELECT MAX(task_order) as max_order FROM task";
        return (int)$this->getOne($sql);
    }

    //{{{ replaceOrder
    /**
     * replaceOrder
     *
     * @access public
     * @todo rewrite me!!
     */
    function replaceOrder($base, $target)
    {
        if (count($base) != count($target)) {
            return false;
        }
        $this->debug = true;
        $this->db->debug = true;

        $sql = "SELECT id, task_order FROM task";
        $rows = $this->getAll($sql);
        $get_id_from_task_order = array();
        foreach ($rows as $row) {
            $get_id_from_task_order[$row['task_order']] = $row['id'];
        }

        $is_commit = true;
        $sql = "UPDATE task SET task_order = ? WHERE id = ?";

        $this->begin();

        foreach ($base as $from) {
            $to = (int)array_shift($target);
            $id = (int)$get_id_from_task_order[$to];

            if ($from != $to) {
                $result = $this->execute($sql, array($from, $id));
                $is_commit = ($is_commit && $result);
            } else {
                continue;
            }
        }

        if ($is_commit) {
            $this->commit();
        } else {
            $this->rollback();
        }

        return $is_commit;

    }
    //}}}

    /**
     * getTaskFromId
     *
     * @access public
     */
    function getTaskFromId($id)
    {
        $id = (int)$id;
        return $this->getRow("SELECT * FROM task WHERE id = ?", array($id));
    }

    /**
     * reNumber
     *
     * @access public
     */
    function reNumber()
    {
        $this->begin();
        $result = true;
        $cnt = 1;
        $list_all = $this->getAll("SELECT * FROM task ORDER BY task_order");
        foreach ($list_all as $row) {
            $sql = "UPDATE task SET task_order = ? WHERE id = ?";
            $result_query = $this->execute($sql, array($cnt, $row['id']));
            $result = ($result && $result_query);
            $cnt++;
        }
        $this->commit();
    }

}
?>
