<?php
/**
 * Php Textfile DB API
 *
 * Copyright 2005 by c-worker.ch
 * http://www.c-worker.ch
 */
/**********************************************************************
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution. 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE. 
***********************************************************************/

/**
 * 0=Debug disabled, 1=Debug enabled
 */
define("TXTDBAPI_DEBUG", 0);

// Even more Debug Infos ?
define("TXTDBAPI_VERBOSE_DEBUG", 0);

/**
 * 0 = Errors are NOT displayed, 1 = Errors are displayed
 */
define("PRINT_ERRORS", 1);

/**
 * 0 = Warnings are NOT displayed, 1 = Warnings are displayed
 */
define("PRINT_WARNINGS", 1);


/**
 * 0=LIKE is case insensitive, 1=LIKE is case sensitive
 */
define("LIKE_CASE_SENSITIVE", 0);

/**
 * 0=ORDER BY is case insensitive, 1=ORDER BY is case sensitive
 */
define("ORDER_CASE_SENSITIVE", 0);

// This constant doesn't limit the max size of a record it's only the assmued size 
// of a record when a table is read for appending. If not a whole record is
// contained in ASSUMED_RECORD_SIZE bytes, the the number of bytes read 
// is increased until a whole record was read. Choosing this value wisely may
// result in a better INSERT performance
define("ASSUMED_RECORD_SIZE", 30);

// General
define("NOT_FOUND",-1);

// File parsing
define("COLUMN_SEPARATOR_CHAR",'#');    // Char which seperates the columns in the table file
                                        // This MUST be a sigle character and NOR a string

// Char to Escape COLUMN_SEPARATOR_CHAR in the Table Files
define("TABLE_FILE_ESCAPE_CHAR",'%');

// Timeouts
define("OPEN_TIMEOUT",10);      // Timeout in seconds to try opening a still locked Table
define("LOCK_TIMEOUT",10);      // Timeout in seconds to try locking a still locked Table
define("LOCKFILE_TIMEOUT",30);  // Timeout for the maximum time a lockfile can exist

// Order Types
define("ORDER_ASC",1);
define("ORDER_DESC",2);

// Join Types
define("JOIN_INNER",1);
define("JOIN_LEFT",2);
define("JOIN_RIGHT",3);

// Row Flags
define("FLAG_KEEP",0x1);

// Column Function Types
define("COL_FUNC_TYPE_SINGLEREC",1);
define("COL_FUNC_TYPE_GROUPING",2);

// Column Types
define("COL_TYPE_INC","inc");
define("COL_TYPE_INT","int");
define("COL_TYPE_STRING","str");

require_once "DB/util.php";
require_once "DB/ResultSet.php";
require_once "DB/SqlParser.php";
require_once "DB/ExpressionParser.php";

/**
 * Database
 *
 * Represents a Database and has functions to execute Sql-Queries on it
 */
class Text_DB {

    /**
     * Text_DB verion
     * @var     string
     * @access  private
     */
    var $version = '0.0.3';

    var $is_debug = false;

    var $dbFolder;

    var $lastInsertId = 0;
    
    var $lastErrorMsgs = array();

    var $table_file_ext = ".txt";
    
    var $table_file_open_mode = ""; // "b" or ""
       
    /**
     * Constructor
     */
    function Text_DB($db_name = "defaultDb/") {             
        $this->dbFolder= $db_name;

        if(!empty($db_name) && last_char($this->dbFolder) != "/") {
            $this->dbFolder .= "/";
        }

        if (defined('TXTDBAPI_DEBUG')) {
            $this->is_debug = TXTDBAPI_DEBUG;
        }
    }
   
    /**
     * updateLastInsertId
     *
     * Insert Id Functions
     */
    function updateLastInsertId($resultSet) { 
        $resultSet->end();
        for($i=0; $i < count($resultSet->colTypes); ++$i) { 
            if($resultSet->colTypes[$i]==COL_TYPE_INC) { 
                $this->lastInsertId=$resultSet->getCurrentValueByNr($i); 
                debug_print("Setting lastInsertId to " . $this->lastInsertId );
                return; 
            } 
        } 
       // if no inc column exists set lastInsertId to 0 
       $this->lastInsertId = 0; 
       debug_print("Setting lastInsertId to " . $this->lastInsertId );
    } 
    
    function getLastInsertId() {
        return $this->lastInsertId;
    }   
    
    /**
     * Table open/close Functions
     *
     * Does open a Table for writing
     */
    function openTableWrite($tableName) {

        $filename = $this->dbFolder . $tableName . $this->table_file_ext;
        
        $fp = fopen($filename, "r+".$this->table_file_open_mode);
        if (!$fp) {
            return 0;
        }
        
        while(!flock($fp, LOCK_EX)) {
            usleep(50);
        }

        return $fp;
    }
    
    // same rules as for openTableWrite()
    // does open a table for appending
    function openTableAppend($tableName) {

        $filename = $this->dbFolder . $tableName . $this->table_file_ext;

        $fp = fopen ($filename, "a".$this->table_file_open_mode);
        if (!$fp) return 0;
        
        while(!flock($fp, LOCK_EX)) {
            usleep(50);
        }

        return $fp;
    }
        
    // Opens a Table for reading
    // This function is used by SELECT
    // returns a FilePointer or false
    function openTableRead($tableName) {
        
        $filename = $this->dbFolder . $tableName . $this->table_file_ext;

        debug_print("Table {$tableName} opened (READ)<br>");

        if(!file_exists($filename)) {
            return 0;
        }

        $fp = fopen($filename, "r".$this->table_file_open_mode);
        if (!$fp) return 0;
        
        while(!flock($fp, LOCK_SH)) {
            usleep(50);
        }

        return $fp;
    }
    
    /**
     * closeTable
     *
     * Close a Table
     *
     * @access private
     * @param resource $fp
     */
    function closeTable($fp) {
        debug_print("Table $fp closed<br>");
        while(!flock($fp, LOCK_UN)) {
            usleep(50);
        }
        return fclose($fp);     
    }
    
    // Reads a Table into a ResultSet
    // Returns a ResultSet or null (the function opens an closes the file itself)
    function readTable($tableName) {
        debug_print("readTable<br>");
        $parser = new ResultSetParser();
        if(!($fd = $this->openTableRead($tableName))) {
            print_error_msg("readTable(): Cannot open Table $tableName");
            return null;
        }
        $rs = $parser->parseResultSetFromFile($fd);
        $this->closeTable($fd);
        return $rs;
    }
        
    // Reads a Table into a ResultSet
    // But only the column names and types + the last record
    // thats usefull for appending record
    function readTableForAppend($tableName) {
        debug_print("readTableForAppend<br>");
        $parser= new ResultSetParser();
        if(!($fd=$this->openTableRead($tableName))) {
            print_error_msg("readTableForAppend(): Cannot open Table $tableName");
            return null;
        }
        $rs=$parser->parseResultSetFromFileForAppend($fd);
        $this->closeTable($fd);
        return $rs;
    }


    // writes the table by using the FilePointer $fd 
    // $fd has to be opened an closed by the caller
    function writeTable($fd, $resultSet) {
        debug_print("writeTable<br>");
        $parser= new ResultSetParser();
        return $parser->parseResultSetIntoFile($fd, $resultSet);
    }
    
    // appends to the table by using the FilePointer $fd 
    // $fd has to be opened an closed by the caller
    function appendToTable($fd, $resultSet,$recordCount) {
        debug_print("appendToTable<br>");
        $parser= new ResultSetParser();
        while($resultSet->getRowCount()>$recordCount) {
            $resultSet->reset();
            $resultSet->next();
            $resultSet->deleteCurrentRow();
        }
        return $parser->parseResultSetIntoFileAppend($fd, $resultSet);
    }
    
    //{{{ executeQuery
    /**
     * executeQuery
     *
     * $sql_query_str is an unparsed SQL Query String
     * Return Values:
     * SELECT Queries: Returns a ResultSet Object or false
     * CREATE TABLE: Returns true or false
     * All other types: Returns the number of rows affected
     *
     * @access public
     */
    function executeQuery($sql_query_str)
    {
        set_error_handler("txtdbapi_error_handler");
        txtdbapi_clear_errors();
        
        debug_print("[executeQuery] Query: {$sql_query_str}<br>", true);
        
        // Parse Query
        $start = getmicrotime();
        $sqlParser = new SqlParser($sql_query_str);
        $sqlQuery = $sqlParser->parseSqlQuery();

        debug_print("parseSqlQuery: " . (getmicrotime() - $start) . " seconds elapsed<br>");

        // free $sqlParser
        unset($sqlParser);
        $sqlParser = '';
        
        // Test Query
        if(!is_object($sqlQuery) || ($sqlQuery->test() === false)) {
            restore_error_handler();
            return false;
        }
        
        $start = getmicrotime();
    
        debug_print("[executeQuery] Parsed Query:<br>", true);
        if($this->is_debug) {
            $sqlQuery->dump();
        }

        // Dispatch
        switch($sqlQuery->type) {

            case "SELECT":
                $rc=$this->executeSelectQuery($sqlQuery);
                break;
            case "INSERT":
                $rc=$this->executeInsertQuery($sqlQuery);
                break;
            case "DELETE":
                $rc=$this->executeDeleteQuery($sqlQuery);
                break;
            case "UPDATE":
                $rc=$this->executeUpdateQuery($sqlQuery);
                break;
            case "CREATE TABLE":
                $rc=$this->executeCreateTableQuery($sqlQuery);
                break;
            case "DROP TABLE":
                $rc=$this->executeDropTableQuery($sqlQuery);
                break;
            case "CREATE DATABASE":
                $rc=$this->executeCreateDatabaseQuery($sqlQuery);
                break;
            case "DROP DATABASE":
                $rc=$this->executeDropDatabaseQuery($sqlQuery);
                break;
            case "LIST TABLES":
                $rc=$this->executeListTablesQuery($sqlQuery);
                break;
            default:
                print_error_msg("Invalid or unsupported Query Type: " . $sqlQuery->type);
                restore_error_handler();
                return false;

        }
        
        if($rc === false) {
            print_error_msg("Query '" . $sql_query_str . "' failed");
        }
        
        debug_print("[executeQuery] Query execution done: " . (getmicrotime() - $start) . " seconds elapsed<br>", true);
        restore_error_handler();        
        return $rc;
    }
    //}}}
    
    /**
     * executeDeleteQuery
     */
    function executeDeleteQuery(&$sqlQuery) {
        
        // Read Table
        $rs = $this->readTable($sqlQuery->tables[0]);
        if(!$rs) {
            print_error_msg("Reading the Table " . $sqlQuery->tables[0] . " failed");
            return false;
        }
        
        $rowsAffected=0;
        
        if(!$sqlQuery->where_expr || $sqlQuery->where_expr=="") {
            $rowsAffected=$rs->getRowCount();
            $rs->deleteAllRows();
        } else {
            // set row ids
            $rId=-1;
            $rs->reset();
            while($rs->next()) 
                $rs->setCurrentRowId(++$rId);
            $rs->reset();
            
            // calc current column count
            $colCount=count($rs->colNames);
            
            // generate additional columns from the WHERE-expression
            $rs->generateAdditionalColumnsFromWhereExpr($sqlQuery->where_expr);
        
            // execute the new single-rec functions
            $rs->executeSingleRecFuncs();
            
            // apply WHERE Expression
            $ep=new ExpressionParser();
            $et=$ep->getExpressionTree($sqlQuery->where_expr);
            $rsFiltered=$et->getFilteredResultSet($rs);
            
            if(!$rsFiltered)
                return 0;           
            
            // Delete rows..
            $rsFiltered->reset();
            while($rsFiltered->next()) {
                $rowId=$rsFiltered->getCurrentRowId();
                $rs->deleteRow($rs->searchRowById($rowId));
            }
            
            $rowsAffected=$rsFiltered->getRowCount();
            
            // Remove columns added from WHERE Expression
            while(count($rs->colNames)>$colCount) {
                $rs->removeColumn(count($rs->colNames)-1);
            }
        }   
        
        // Open Table
        $fp=$this->openTableWrite($sqlQuery->tables[0]);
        if(!$fp) {
            print_error_msg("Open the Table " . $sqlQuery->tables[0] . " (for WRITE) failed");
            return false;
        }
        
        // Write Table
        $this->writeTable($fp,$rs);
        $this->closeTable($fp);
        return $rowsAffected;
    }

    //{{{ executeInsertQuery
    /**
     * executeInsertQuery
     *
     * returns the affected Row count or false
     */
    function executeInsertQuery(&$sqlQuery) {
        
        // Read Table
        $rs=$this->readTableForAppend($sqlQuery->tables[0]);
        
        if(TXTDBAPI_VERBOSE_DEBUG) {
            echo "executeInsertQuery(): Last Record read for appending:<br>";
            $rs->dump();
            echo "<br>";
        }
        
        if(!$rs) {
            print_error_msg("Reading the Table " . $sqlQuery->tables[0] . " failed");
            return false;
        }
        
        // Open Table
        $fp=$this->openTableAppend($sqlQuery->tables[0]);
        if(!$fp) {
            print_error_msg("Open the Table " . $sqlQuery->tables[0] . " (for APPEND) failed");
            return false;
        }
        
        // a INSERT INTO table () VALUES () query: just write the default values
        if(count($sqlQuery->colNames)==0 && count($sqlQuery->fieldValues)==0) {
                $rs->append();
                $this->updateLastInsertId($rs);
                $this->appendToTable($fp,$rs,1);
                $this->closeTable($fp);
                return 1; // Error Handling ??
        }
                
        // execute functions on the values
        $colName="";
        $colTable="";
        $colFunc="";
        
        for($i=0;$i<count($sqlQuery->fieldValues);++$i) {
            split_full_colname($sqlQuery->fieldValues[$i],$colName,$colTable,$colFunc);
            if($colFunc) {
                if($colName && has_quotes($colName))  {
                    remove_quotes($colName);
                }
                $sqlQuery->fieldValues[$i]=execFunc($colFunc, $colName);
            } else {
                if(has_quotes($sqlQuery->fieldValues[$i])) {
                    remove_quotes($sqlQuery->fieldValues[$i]);
                }
            }
        }
                
        $rc=true;
        switch(count($sqlQuery->colNames)) {
            case 0:
                $rs->appendRow($sqlQuery->fieldValues);
                $this->updateLastInsertId($rs);
                $this->appendToTable($fp,$rs,1);
                $this->closeTable($fp);
                return $rc;
                break;
            default:
                $rs->append();
                for($i=0;$i<count($sqlQuery->colNames);++$i) {
                    if(!$rs->setCurrentValueByName($sqlQuery->colNames[$i],$sqlQuery->fieldValues[$i])) {
                        $rc=false;
                        break;
                    }
                }
                if($rc) {
                    $this->updateLastInsertId($rs);
                    $this->appendToTable($fp,$rs,1);
                }
                $this->closeTable($fp);
                return $rc;
                break;
        }
    }
    //}}}
    
    //{{{ executeUpdateQuery
    /**
     * executeUpdateQuery
     *
     * returns the affected Row count or false
     */
    function executeUpdateQuery(&$sqlQuery) {
        
        $rs=$this->readTable($sqlQuery->tables[0]);
        if(!$rs) {
            print_error_msg("Reading the Table " . $sqlQuery->tables[0] . " failed");
            return false;
        }
        
        // calc original column count
        $colCount=count($rs->colNames);

        $rs->generateAdditionalColumnsFromArray($sqlQuery->fieldValues);

        if(txtdbapi_error_occurred())
            return false;

        $rs->executeSingleRecFuncs();

        // check if there are wrong functions
        for($i=0;$i<count($rs->colFuncs);++$i) {
            if($rs->colFuncs[$i] && !$rs->colFuncsExecuted[$i]) {
                print_error_msg("Function '" . $rs->colFuncs[$i]  . "' not supported in UPDATE statements");
                return false;
            }
        }
        
        if($this->is_debug) {
            debug_print("[executeUpdateQuery] ResultSet dump after generating columns:<br>", true);
            $rs->dump();
        }
        
        // No where_expr ? update all
        if( (!isset($sqlQuery->where_expr)) || (!$sqlQuery->where_expr) ) {
            // update 
            $rs->reset();
            while($rs->next()) {
                for($i=0;$i<count($sqlQuery->colNames);++$i) {
                    $rc=$rs->setCurrentValueByName($sqlQuery->colNames[$i],
                        $rs->getCurrentValueByName($sqlQuery->fieldValues[$i]));
                    if(!$rc)  {
                        return false;
                    }
                }
            }
            
            // Remove columns added from WHERE Expression
            while(count($rs->colNames)>$colCount) {
                $rs->removeColumn(count($rs->colNames)-1);
            }
            
            // Open Table
            $fp=$this->openTableWrite($sqlQuery->tables[0]);
            if(!$fp) {
                print_error_msg("Open the Table " . $sqlQuery->tables[0] . " (for WRITE) failed");
                return false;
            }
            // Write Table
            $this->writeTable($fp,$rs);
            $this->closeTable($fp);
            return $rs->getRowCount();

        } else {
            // set row id's
            $rs->reset();
            $rId = -1;
            while($rs->next()) {
                $rs->setCurrentRowId(++$rId);
            }
            
            // generate additional columns from the WHERE-expression
            $rs->generateAdditionalColumnsFromWhereExpr($sqlQuery->where_expr);
        
            // execute the new single-rec functions
            $rs->executeSingleRecFuncs();
            
            // create a copy 
            $rsFiltered=$rs;

            // filter by where expression
            $ep=new ExpressionParser();
            $et=$ep->getExpressionTree($sqlQuery->where_expr);
            $rsFiltered=$et->getFilteredResultSet($rsFiltered);
            
            if($rsFiltered->getRowCount()<1)
                return 0;
                
            // update 
            $rsFiltered->reset();
            while($rsFiltered->next()) {
                for($i=0;$i<count($sqlQuery->colNames);++$i) {
                    $rc=$rsFiltered->setCurrentValueByName($sqlQuery->colNames[$i],
                        $rsFiltered->getCurrentValueByName($sqlQuery->fieldValues[$i]));
                    if(!$rc)  {
                        return false;
                    }
                    
                }
            }
                        
            // put filtered part back in the original ResultSet
            $rowNr=0;
            $putBack=0;
            $rs->reset();
            $rsFiltered->reset();
            while($rs->next()) {
                if(($rowNr=$rsFiltered->searchRowById($rs->getCurrentRowId())) !=NOT_FOUND) {
                    $rs->setCurrentValues($rsFiltered->getValues($rowNr));
                    $putBack++;
                }
            }
            if($putBack<$rsFiltered->getRowCount()) {
                print_error_msg("UPDATE: Could not put Back all filtered Values");
                return false;
            }
            
            
            // Remove columns added from WHERE Expression
            while(count($rs->colNames)>$colCount) {
                $rs->removeColumn(count($rs->colNames)-1);
            }
            
            // Open Table
            $fp = $this->openTableWrite($sqlQuery->tables[0]);
            if(!$fp) {
                print_error_msg("Open the Table " . $sqlQuery->tables[0] . " (for WRITE) failed");
                return false;
            }
            $this->writeTable($fp,$rs);
            $this->closeTable($fp);
            return $rsFiltered->getRowCount();
        }
    }
    //}}}
    
    //{{{ executeCreateTableQuery
    /**
     * executeCreateTableQuery
     *
     * executes a SQL CREATE TABLE Statement
     * param: SqlQuery Object
     * returns True or False
     */
    function executeCreateTableQuery(&$sqlQuery) {

        clearstatcache();
        $filename = $this->dbFolder . $sqlQuery->tables[0] . $this->table_file_ext;

        // checks
        if (!is_string($sqlQuery->tables[0])) {
            print_error_msg("Invalid Table " . $sqlQuery->tables[0]);
            return false;
        }

        if (file_exists($filename)) {
            print_error_msg("Table " . $sqlQuery->tables[0] . " allready exists");
            return false;
        }

        $col_type_count = count($sqlQuery->colTypes);

        if(count($sqlQuery->colNames) != $col_type_count) {
            print_error_msg("There's not a type defined for each column");
            return false;
        }
        
        for($i = 0; $i < $col_type_count; ++$i) {

            $sqlQuery->colTypes[$i] = strtolower($sqlQuery->colTypes[$i]);

            if (!in_array($sqlQuery->colTypes[$i], array(COL_TYPE_INT, COL_TYPE_STRING, COL_TYPE_INC))) {
                print_error_msg("Column Type {$sqlQuery->colTypes[$i]} not supported");
                return false;
            }
        }

        $fp = fopen($filename, "w".$this->table_file_open_mode);
        
        $rsParser = new ResultSetParser();

        $data = $rsParser->parseLineFromRow($sqlQuery->colNames);
        $data.= "\n";
        $data.= $rsParser->parseLineFromRow($sqlQuery->colTypes);
        $data.= "\n";
        $data.= $rsParser->parseLineFromRow($sqlQuery->fieldValues);

        fwrite($fp, $data);
        fclose($fp);

        chmod($filename, 0666);

        return true;    
    }
    //}}}
    
    // executes a SQL DROP TABLE Statement 
    // param: SqlQuery Object
    // returns True or False
    function executeDropTableQuery(&$sqlQuery) {
        clearstatcache();
        if(!isset($sqlQuery->colNames[0]))
            return false;
        
        for($i=0;$i<count($sqlQuery->colNames);++$i) {
            $rc=unlink($this->dbFolder. $sqlQuery->colNames[$i]  . $this->table_file_ext);
            if(!$rc) {
                print_error_msg("DROP TABLE " . $sqlQuery->colNames[$i] . " failed");
                return false;
            }
        }
        return true;    
    }
    
    /***********************************
            List Tables Query
    ************************************/
    // executes a LIST TABLES Statement 
    // param: SqlQuery Object
    // returns: A ResultSet Object with a single column "table"
    function executeListTablesQuery(&$sqlQuery)
    {

        $rs = new ResultSet();
        $rs->colNames=array("table");
        $rs->colAliases=array("");
        $rs->colTables=array("");
        $rs->colTableAliases=array("");
        $rs->colTypes=array(COL_TYPE_STRING);
        $rs->colDefaultValues=array("");
        $rs->colFuncs=array("");
        $rs->colFuncsExecuted=array(false);
        
        $handle = opendir($this->dbFolder);
        
        $rs->reset();
        while ($file = readdir ($handle)) { 
            if ($file != "." && $file != ".." && is_file($this->dbFolder . $file)) { 
                $rs->appendRow(array(substr($file,0,strlen($file)-strlen($this->table_file_ext))));
            } 
        }
        
        // apply WHERE Statement
        if($sqlQuery->where_expr) {
            $ep=new ExpressionParser();
            $et=$ep->getExpressionTree($sqlQuery->where_expr);
            $rs=$et->getFilteredResultSet($rs);
        } 
            
        // Order ResultSet
        if(count($sqlQuery->orderColumns)>0) {
            $rs->orderRows($sqlQuery->orderColumns,$sqlQuery->orderTypes);
        }

        // Group ResultSet (process GROUP BY)
        if(count($sqlQuery->groupColumns)>0) {
            $rs = $rs->groupRows($sqlQuery->groupColumns, $sqlQuery->limit);
        }

        // Apply Limit      
        $rs->reset();
        $rs = $rs->limitResultSet($sqlQuery->limit);
        
        
        closedir($handle); 
        $rs->reset();
        return $rs;
    }
    
    /***********************************
            Create Database Query
    ************************************/
    // executes a SQL CREATE DATABASE Statement 
    // param: SqlQuery Object
    // returns True or False
    function executeCreateDatabaseQuery(&$sqlQuery) {

        clearstatcache();
        if (!isset($sqlQuery->colNames[0])) {
            return false;
        }

        $directory = dirname($this->dbFolder) . '/' . $sqlQuery->colNames[0];

        if (!is_dir($directory)) {
            $rc = mkdir($directory, 0777);
            if(!$rc) {
                print_error_msg("Cannot create Database : mkdir failed [{$sqlQuery->colNames[0]}]");
                return false;
            }
        }
        return true;
    }
    
    //{{{ executeDropDatabaseQuery
    /**
     * executeDropDatabaseQuery
     *
     * executes a SQL DROP DATABASE Statement 
     * param: SqlQuery Object
     * returns True or False
     */
    function executeDropDatabaseQuery(&$sqlQuery) {
        clearstatcache();
        if(!isset($sqlQuery->colNames[0]))
            return false;
        $directory=$this->dbFolder . $sqlQuery->colNames[0];
        
        // delete all tables
        $dirHandle=opendir($directory); 
            while ($file = readdir ($dirHandle)) { 
                if ($file != "." && $file != ".." && is_file($directory . "/" . $file)) { 
                if(!($rc=unlink($directory . "/" . $file))) {
                    print_error_msg("Cannot drop Database: Deleting the table $file failed");
                }
                debug_print($file . "<br>");
            } 
        }
        closedir($dirHandle); 

        
        $rc=rmdir ($directory);
        if(!$rc) {
            print_error_msg("Cannot drop Database " . $sqlQuery->colNames[0]);
            return false;
        }
        return true;    
    }
    //}}}
    
    /***********************************
            Select Query
    ************************************/
    // executes a SQL SELECT STATEMENT and returns a ResultSet 
    // param: SqlQuery Object
    function executeSelectQuery(&$sqlQuery) {

        global $g_sqlGroupingFuncs;
        global $g_sqlSingleRecFuncs;

        $resultSets=array();        
        
        // create a copy
        $aliases=$sqlQuery->colAliases;
        $funcs=$sqlQuery->colFuncs;
            
        // Read all Tables
        for($i=0;$i<count($sqlQuery->tables);++$i) {
            debug_print("<br>[executeSelectQuery] Reading table " . $sqlQuery->tables[$i] ."<br>", true); 
            if(!($resultSets[$i]=$this->readTable($sqlQuery->tables[$i]))) {
                print_error_msg("Reading Table " . $sqlQuery->tables[$i]. " failed");
                return false;
            }
            $resultSets[$i]->setColumnTableForAll($sqlQuery->tables[$i]);
            $resultSets[$i]->setColumnTableAliasForAll($sqlQuery->tableAliases[$i]);
            
            // set all functions to the ResultSet of the current table
            // if table and column name matches
            debug_print("[executeSelectQuery] Setting functions for the current table:<br>", true);
            for($j=0;$j<count($funcs);++$j) {
                if(!$funcs[$j] || !$sqlQuery->colNames[$j])
                    continue;
                                
                if($sqlQuery->colTables[$j]==$sqlQuery->tables[$i] ||
                   (strlen($sqlQuery->tableAliases[$i]) > 0 && ($sqlQuery->colTables[$j]==$sqlQuery->tableAliases[$i]))) {
                    $colNr=$resultSets[$i]->findColNrByAttrs($sqlQuery->colNames[$j],$sqlQuery->colTables[$j],"");
                    
                    if($colNr==NOT_FOUND) {
                        continue;
                    }                   
                    
                    // create a new column for each function
                    $resultSets[$i]->addColumn($sqlQuery->colNames[$j],$sqlQuery->colAliases[$j],$sqlQuery->colTables[$j],"","str","",$funcs[$j],"",true);
                    $funcs[$j]="";
                }
                
            }
            
            // set all aliases where table, column name and function matches
            debug_print("[executeSelectQuery] Setting aliases for the current table:<br>", true);
            for($j=0;$j<count($aliases);++$j) {
                if(!$aliases[$j])
                    continue;
                
                if($sqlQuery->colTables[$j]==$sqlQuery->tables[$i] ||
                   $sqlQuery->colTables[$j]==$sqlQuery->tableAliases[$i]) { 
                    if(($colNr=$resultSets[$i]->findColNrByAttrs($sqlQuery->colNames[$j],$sqlQuery->colTables[$j],$sqlQuery->colFuncs[$j])) != NOT_FOUND) {
                        $resultSets[$i]->setColumnAlias($colNr,$aliases[$j]);
                        $aliases[$j]="";                
                    }
                }
            }
            
            if($this->is_debug) {
                debug_print("<br>[executeSelectQuery] Dump of Table $i (" . $sqlQuery->tables[$i] . "):<br>", true);
                $resultSets[$i]->dump();
            }
        }
        
        // set remaining functions to the ResultSet where column name matches
        debug_print("[executeSelectQuery] Setting remaining functions where column name matches:<br>", true);
        for($i=0;$i<count($resultSets);++$i) {
            for($j=0;$j<count($funcs);++$j) {
                if(!$funcs[$j] || !$sqlQuery->colNames[$j])
                    continue;
                                
                $colNr=$resultSets[$i]->findColNrByAttrs($sqlQuery->colNames[$j],"","");
                if($colNr==NOT_FOUND) {
                    // 'text' or 123 ? => add column
                    if(! (is_numeric($sqlQuery->colNames[$j]) || has_quotes($sqlQuery->colNames[$j]))) {
                        continue;
                    }
                    debug_print("Adding function with quoted string or number paremeter!<br>");
                }
                    
                // create a new column for each function
                $resultSets[$i]->addColumn($sqlQuery->colNames[$j],$sqlQuery->colAliases[$j],$sqlQuery->colTables[$j],"","str","",$funcs[$j],"",true);
                $funcs[$j]="";

            }
        }
        
        // set remaining aliases where column name and function matches
        debug_print("[executeSelectQuery] Setting remaining aliases where column name and function matches:<br>", true);
        for($i=0;$i<count($resultSets);++$i) {
            for($j=0;$j<count($aliases);++$j) {
                if(!$aliases[$j])
                    continue;
                if(($colNr=$resultSets[$i]->findColNrByAttrs($sqlQuery->colNames[$j],"",$sqlQuery->colFuncs[$j])) !=NOT_FOUND) {
                    $resultSets[$i]->setColumnAlias($colNr,$aliases[$j]);
                    $aliases[$j]="";
                }
            }
        }

        debug_print("[executeSelectQuery] Executing single-rec functions (on the separate ResultSet's):<br>", true);
        // execute single-rec functions (on the separate ResultSet's)
        for($i=0;$i<count($resultSets);++$i) {
            $resultSets[$i]->executeSingleRecFuncs();
        }
        
        // A query without tables ? => make a dummy ResultSet
        $dummyResultSet=false;
        if(count($sqlQuery->tables)==0) {
            $dummyResultSet=true;
            $rsMaster=new ResultSet();  
            $rsMaster->addColumn("(dummy)", "(dummy)", "(dummy)", "(dummy)", "str", "(dummy)", "", "", true);
            $rsMaster->append();
        
        // else: real ResultSet
        } else {
            $dummyResultSet=false;
            
            /////
            // Perform JOIN's
            foreach($sqlQuery->joins as $join) {
                if(!isset($rsMaster)) {
                    $rsMaster=$resultSets[$join->leftTableIndex];
                    $resultSets[$join->leftTableIndex]=0;
                }
                
                // INNER => Nothing special
                if($join->type == JOIN_INNER) {
                    $rsMaster=$rsMaster->joinWithResultSet($resultSets[$join->rightTableIndex]);
                    $ep=new ExpressionParser();
                    $et=$ep->getExpressionTree($join->expr);
                    $rsMaster=$et->getFilteredResultSet($rsMaster);
                    $resultSets[$join->rightTableIndex]=0;
                    
                // LEFT OUTER JOIN => keep all of the left
                } else if($join->type == JOIN_LEFT) {
                    // generate new row id's for the left
                    $rsMaster->generateRowIds();
                    // copy the left
                    $copy=$rsMaster->copy();
                    $rsMaster=$rsMaster->joinWithResultSet($resultSets[$join->rightTableIndex],0);
                    $ep=new ExpressionParser();
                    $et=$ep->getExpressionTree($join->expr);
                    $rsMaster=$et->getFilteredResultSet($rsMaster);
                    $resultSets[$join->rightTableIndex]=0;
                    // add back missing
                    $rsMaster->addMissingRows($copy);
                    
                    
                // RIGHT OUTER JOIN => keep all of the right
                } else if($join->type == JOIN_RIGHT) {
                    // generate new row id's for the right
                    $resultSets[$join->rightTableIndex]->generateRowIds();
                    // copy the right
                    $copy=$resultSets[$join->rightTableIndex]->copy();
                    $rsMaster=$rsMaster->joinWithResultSet($resultSets[$join->rightTableIndex],1);
                    $ep=new ExpressionParser();
                    $et=$ep->getExpressionTree($join->expr);
                    $rsMaster=$et->getFilteredResultSet($rsMaster);
                    $resultSets[$join->rightTableIndex]=0;              
                    // add back missing
                    $rsMaster->addMissingRows($copy);
                }
            }
        
            // join the remaining ResultSet's
            if(!isset($rsMaster)) {
                $rsMaster=$resultSets[0];
                $i=1;
            } else {
                $i=0;
            }
            for(;$i<count($resultSets);++$i) {
                if($resultSets[$i] != 0) {
                    $rsMaster=$rsMaster->joinWithResultSet($resultSets[$i]);
                }
            }
        }
        
        // from here we only work with $rsMaster and can free the other ResultSet's
        unset($resultSets);
        $resultSets="";
        
        // generate additional columns for COUNT(*) functions
        debug_print("[executeSelectQuery] Adding COUNT(*) functions<br>", true);
        for($i=0;$i<count($funcs);++$i) {
            if($funcs[$i] && $sqlQuery->colNames[$i]=="*") {
                $rsMaster->addColumn($sqlQuery->colNames[$i],$sqlQuery->colAliases[$i],$sqlQuery->colTables[$i],"","int","",$funcs[$i],"","");
                $funcs[$i]="";
            }
        }
        
        // generate additional columns for the remaining functions (functions without params)
        for($i=0;$i<count($funcs);++$i) {
            if($funcs[$i]) {
                $rsMaster->addColumn($sqlQuery->colNames[$i],$sqlQuery->colAliases[$i],"","","str","",$funcs[$i],execFunc($funcs[$i],""));
            }
        }

        // generate additional columns from the WHERE-expression
        $rsMaster->generateAdditionalColumnsFromWhereExpr($sqlQuery->where_expr);
        
        // generate additional columns from ORDER BY
        $rsMaster->generateAdditionalColumnsFromArray($sqlQuery->orderColumns);
        
        // generate additional columns from GROUP BY
        $rsMaster->generateAdditionalColumnsFromArray($sqlQuery->groupColumns);
        
        // execute the new single-rec functions (on the Master ResultSet)
        $rsMaster->executeSingleRecFuncs();
        
        // generate new row id's 
        $rsMaster->generateRowIds();
            
        debug_print("<br>[executeSelectQuery] Master ResultSet:</b><br>", true);
        if($this->is_debug) {
            $rsMaster->dump();
        }
        
        // apply WHERE expression
        if($sqlQuery->where_expr) {
            $ep=new ExpressionParser();
            $et=$ep->getExpressionTree($sqlQuery->where_expr);
            $rsMaster=$et->getFilteredResultSet($rsMaster);
        } 
        // free $ep
        unset($ep);
        $ep="";
        
        // stop if the WHERE expression failed
        if(txtdbapi_error_occurred()) {
            return false;
        }
        
        // check if we can use some optimization 
        // (use the limit in group by, but only if there are no grouping function
        // in the groupRows. To be able to do this we must order before grouping)
        $optimizedPath=true;
        if(!$sqlQuery->limit || !$sqlQuery->orderColumns) {
            $optimizedPath=false;
        } else {
            for($i=0;$i<count($sqlQuery->colFuncs);++$i) {
                if(in_array($sqlQuery->colFuncs[$i],$g_sqlGroupingFuncs)) {
                    $optimizedPath=false;
                    break;
                }
            }               
        }
        if($optimizedPath) {
            debug_print("[executeSelectQuery] Using optimized path!<br>", true);
        } else {
            debug_print("[executeSelectQuery] Using normal path!<br>", true);
        }
        
        // Order ResultSet (if optimizedPath)
        if($optimizedPath) {    
            debug_print("[executeSelectQuery] Calling orderRows() (optimized path)..<br>", true);
            if(count($sqlQuery->orderColumns)>0) {
                $rsMaster->orderRows($sqlQuery->orderColumns,$sqlQuery->orderTypes);
            }
        }
        
        // Group ResultSet (process GROUP BY)
        $numGroupingFuncs=0;
        for($i=0;$i<count($sqlQuery->colFuncs);++$i) {
            if($sqlQuery->colFuncs[$i] && in_array($sqlQuery->colFuncs[$i],$g_sqlGroupingFuncs)) {
                ++$numGroupingFuncs;
                break;
            }
        }
        if($numGroupingFuncs>0 || count($sqlQuery->groupColumns)>0) {
            debug_print("[executeSelectQuery] Calling groupRows()..<br>", true);
            $rsMaster = $rsMaster->groupRows($sqlQuery,$optimizedPath);
        }
        
        // Order ResultSet (if NOT optimizedPath)
        if(!$optimizedPath) {   
            debug_print("[executeSelectQuery] Calling orderRows() (normal path)..<br>", true);
            if(count($sqlQuery->orderColumns)>0) {
                $rsMaster->orderRows($sqlQuery->orderColumns,$sqlQuery->orderTypes);
            }
        }

        // add direct value columns
        debug_print("[executeSelectQuery] Adding direct value columns..<br>", true);
        for($i=0;$i<count($sqlQuery->colNames);++$i) {
            if($sqlQuery->colNames[$i] && (is_numeric($sqlQuery->colNames[$i]) || has_quotes($sqlQuery->colNames[$i])) &&
               !$sqlQuery->colTables[$i] && !$sqlQuery->colFuncs[$i] &&
               $rsMaster->findColNrByAttrs($sqlQuery->colNames[$i],"","")==NOT_FOUND) {
                $value=$sqlQuery->colNames[$i];
                if(has_quotes($value)) {
                    remove_quotes($value);
                }
                $rsMaster->addColumn($sqlQuery->colNames[$i],$sqlQuery->colAliases[$i],"","","str","","",$value,true);
            }
        }
        
        // return only the requested columns
        debug_print("[executeSelectQuery] Removing unwanted columns...<br>", true);
        $rsMaster->filterByColumnNamesInSqlQuery($sqlQuery);
        
        
        // order columns (not their data)
        debug_print("[executeSelectQuery] Ordering columns (amog themself)...<br>", true);
        if(!$rsMaster->orderColumnsBySqlQuery($sqlQuery)) {
            print_error_msg("Ordering the Columns (themself) failed");
            return false;
        }   
        
        // process DISTINCT
        if($sqlQuery->distinct == 1) {
            $rsMaster = $rsMaster->makeDistinct($sqlQuery->limit);
        }
        
        // Apply Limit      
        $rsMaster->reset();
        $rsMaster = $rsMaster->limitResultSet($sqlQuery->limit);
        verbose_debug_print ("<br>Limited ResultSet:<br>");
        if(TXTDBAPI_VERBOSE_DEBUG) $rsMaster->dump();

        
        $rsMaster->reset();
        return $rsMaster;
    }
}
?>
