<?php
/**
 * Php Textfile DB API
 * Copyright 2005 by c-worker.ch
 * http://www.c-worker.ch
 */
/**********************************************************************
Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met: 
Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer. 
Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution. 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
THE POSSIBILITY OF SUCH DAMAGE.
***********************************************************************/

$GLOBALS['g_txtdbapi_errors'] = array();

function debug_print($str, $strong = false) {
    if(!TXTDBAPI_DEBUG) {
        return false;
    }

    if ($strong === true) {
        echo "<strong>{$str}</strong>";
    } else {
        echo $str;
    }
}

function verbose_debug_print($str) {
    if(TXTDBAPI_VERBOSE_DEBUG) {
        echo $str;
    }
}

function last_char($string) {
    $len = strlen($string);
    if($len < 1) {
        return '';
    }
    return $string{strlen($string)-1};
}

function remove_last_char(&$string) {
    $string = substr($string, 0, strlen($string) - 1);
}

/**
 * substr_right
 *
 * returns $length chars from the right side of $string
 */
function substr_right($string, $length) {
    return substr($string, strlen($string)-$length);
}

function create_array_fill($size, $value) {
    $arr = array();
    for($i = 0; $i < $size; ++$i) {
        $arr[] = $value;
    }
    return $arr;
}

function is_0($var) {
    return (is_int($var) && !$var);
}
// _ at the front, cause is_null exists
function _is_null($var) {
    return (!isset($var)) ;
}
function is_empty_str($var) {
    return (is_string($var) && $var=="");
}

/**
 * split_full_colname
 *
 * splits a full column name into its subparts (name, table, function)
 * return true or false on error
 */
function split_full_colname($fullColName, &$colName, &$colTable, &$colFunc) {
    
    $colName = "";
    $colTable= "";
    $colFunc = "";
    
    // direct value ?
    if(is_numeric($fullColName) || has_quotes($fullColName)) {
        $colName = trim($fullColName);
        return true;
    }
    
    $pos = strpos($fullColName, "(");
    if ($pos !== false) {
        $colFunc=strtoupper(trim(substr($fullColName,0,$pos)));
        $fullColName=substr($fullColName,$pos+1);
    }
    
    $pos = strpos($fullColName,".");
    if (($pos !== false) && $colFunc != "EVAL") {
        $colTable = substr($fullColName,0,$pos);
        $colName = substr($fullColName,$pos+1);
    }  else {
        $colName=$fullColName;
    }

    $colName=trim($colName);
    if($colFunc) {
        if(last_char($colName) == ")") {
            remove_last_char($colName);
        } else {
            print_error_msg(") expected after {$colName}! at Line:" . __LINE__);
            return false;
        }
    }

    $colName = trim($colName);
    $colTable= trim($colTable);

    return true;
}

function execFunc($func, $param) {
    switch($func) {
        case "MD5":
            return md5($param);
        case "NOW":
            return doFuncNOW();
        case "UNIX_TIMESTAMP":
            return doFuncUNIX_TIMESTAMP();
        case "ABS":
            return abs($param);
        case "LCASE":
        case "LOWER":
            return doFuncLCASE($param);
        case "UCASE":
        case "UPPER":
            return doFuncUCASE($param);
        case "INC":
            return $param+1;
        case "DEC":
            return $param-1;
        case "EVAL":
            if(is_empty_str($param))
                return 0;
            return eval("return ($param);");
        default:
            print_error_msg("function '$func' not supported!");
            return $param;
    }
    return $col;
}

function doFuncNOW() {
    return date("Y-m-d H:i:s",get_static_timestamp());
}

function doFuncUNIX_TIMESTAMP() {
    return get_static_timestamp();
}

function doFuncLCASE($param) {
    return strtolower($param);
}

function doFuncUCASE($param) {
    return strtoupper($param);
}

function execGroupFunc($func, $params) {
    
    switch($func) {
        //case "":
        //  return $params[0];
        case "MAX":
            return doFuncMAX($params);
        case "MIN":
            return doFuncMIN($params);
        case "COUNT":
            return doFuncCOUNT($params);
        case "SUM":
            return doFuncSUM($params);
        case "AVG":
            return doFuncAVG($params);
            
        default:
            print_error_msg("Function '$func' not supported!!!");
    }
    return $col;
}

function doFuncMAX($params) {
    $maxVal=$params[0];
    for($i=1;$i<count($params);++$i) {
        $maxVal=max($maxVal,$params[$i]);
    }
    return $maxVal;
}

function doFuncMIN($params) {
    $minVal=$params[0];
    for($i=1;$i<count($params);++$i) {
        $minVal=min($minVal,$params[$i]);
    }
    return $minVal;
}

function doFuncCOUNT($params) {
    return count($params);
}

function doFuncSUM($params) {
    $sum=0;
    for($i=0;$i<count($params);++$i) {
        $sum+=$params[$i];
    }
    return $sum;
}

function doFuncAVG($params) {
    $sum=doFuncSUM($params);
    return $sum / count($params);
}

/***********************************
        Error Functions
************************************/
function print_error_msg($text, $nr=-1) {
    global $g_txtdbapi_errors;
    
    $g_txtdbapi_errors[]=$text;

    if (!PRINT_ERRORS)
        return;

    if ($nr==-1) {
        echo "<br><b>Php-Txt-Db-Access Error:</b><br>";
    } else {
        echo "<br> Php-Txt-Db-Access Error Nr {$nr}<br>";
    }

    echo $text . "<br>";    
}

function print_warning_msg($text, $nr=-1) {
    if(!PRINT_WARNINGS)
        return;
        
    if ($nr == -1) {
        echo "<br><b>Php-Txt-Db-Access Warning:</b><br>";
    } else {
        echo "<br> Php-Txt-Db-Access Warning Nr $nr:<br>";
    }
    
    echo "{$text}<br>";    
}

// returns true if errors occurred
function txtdbapi_error_occurred() {
    global $g_txtdbapi_errors;
    return (count($g_txtdbapi_errors)>0);
}

//using pear driver
function txtdbapi_get_last_error() {
    global $g_txtdbapi_errors;
    if(!txtdbapi_error_occurred())
        return "";
    return array_pop($g_txtdbapi_errors);
}

function txtdbapi_clear_errors()
{
    $GLOBAL['g_txtdbapi_errors'] = array();
}

// error handler function
function txtdbapi_error_handler ($errno, $errstr, $errfile, $errline) {
    $prefix="PHP Error: ";
    switch ($errno) {
        case E_USER_ERROR:
            print_error_msg($prefix . "FATAL [$errno] $errstr [Line: ".$errline."] [File: ". $errfile . "]");
            break;
        default:
            print_error_msg($prefix . "[$errno] $errstr [Line: ".$errline."] [File: ". $errfile . "]");
            break;
    }
}

function has_quotes($str) {
    if (is_empty_str($str)) {
        return false;
    }
    return ($str[0]=="'" || $str[0]=="\"") && (last_char($str)=="'" || last_char($str)=="\"");
}

function remove_quotes(&$str) {
    $str = substr($str, 1);
    remove_last_char($str);
}

function getmicrotime(){ 
    list($usec, $sec) = explode(" ",microtime()); 
    return ((float)$usec + (float)$sec); 
} 

/**
 * get_static_timestamp
 *
 * ensures that all timestamp requests of
 * one execution have the same time
 */
function get_static_timestamp() {
    static $t = 0;
    if ($t == 0) {
        $t = time();
    }
    return $t;
}
?>
