<?php
/**
 * adodb-textdb,inc.php
 *
 * V4.92a 29 Aug 2006  (c) 2000-2006 John Lim (jlim#natsoft.com.my). All rights reserved.
 * Released under both BSD license and Lesser GPL library license. 
 * Whenever there is any discrepancy between the two licenses, 
 * the BSD license will take precedence.
 *
 *  Install Instructions:
 *  ====================
 *  1. Place this in adodb/drivers
 *  2. Rename the file, remove the .txt prefix.
 */

error_reporting(E_ALL);

// security - hide paths
if (!defined('ADODB_DIR')) {
    exit();
}

require_once 'Text/DB.php';

class ADODB_textdb extends ADOConnection {

    var $databaseType = "textdb";
    var $replaceQuote = ""; // string to use to replace quotes
    var $concat_operator='||';
    var $_errorNo = 0;
    var $hasLimit = true;   
    var $hasInsertID = true;        /// supports autoincrement ID?
    var $hasAffectedRows = true;    /// supports affected rows for update/delete?
    //var $metaTablesSQL = "SELECT name FROM sqlite_master WHERE type='table' ORDER BY name";
    var $sysDate = "adodb_date('Y-m-d')";
    var $sysTimeStamp = "adodb_date('Y-m-d H:i:s')";
    var $fmtTimeStamp = "'Y-m-d H:i:s'";
    
    function ADODB_textdb() 
    {
    }
    
    function ServerInfo()
    {
        /*
        $arr['version'] = sqlite_libversion();
        $arr['description'] = 'Text_DB ';
        $arr['encoding'] = sqlite_libencoding();
        return $arr;
         */
    }
    
    //{{{ BeginTrans
    /**
     * BeginTrans
     */
    function BeginTrans()
    {     
         return true;
    }
    //}}}
    
    //{{{ CommitTrans
    function CommitTrans($ok = true) 
    { 
         return true;
    }
    //}}}
    
    //{{{ RollbackTrans
    function RollbackTrans()
    {
         return true;
    }
    //}}}
    
    // mark newnham
    function &MetaColumns($tab)
    {
      global $ADODB_FETCH_MODE;
      $false = false;
      $save = $ADODB_FETCH_MODE;
      $ADODB_FETCH_MODE = ADODB_FETCH_ASSOC;
      if ($this->fetchMode !== false) $savem = $this->SetFetchMode(false);
      $rs = $this->Execute("PRAGMA table_info('$tab')");
      if (isset($savem)) $this->SetFetchMode($savem);
      if (!$rs) {
        $ADODB_FETCH_MODE = $save; 
        return $false;
      }
      $arr = array();
      while ($r = $rs->FetchRow()) {
        $type = explode('(',$r['type']);
        $size = '';
        if (sizeof($type)==2)
        $size = trim($type[1],')');
        $fn = strtoupper($r['name']);
        $fld = new ADOFieldObject;
        $fld->name = $r['name'];
        $fld->type = $type[0];
        $fld->max_length = $size;
        $fld->not_null = $r['notnull'];
        $fld->default_value = $r['dflt_value'];
        $fld->scale = 0;
        if ($save == ADODB_FETCH_NUM) $arr[] = $fld;    
        else $arr[strtoupper($fld->name)] = $fld;
      }
      $rs->Close();
      $ADODB_FETCH_MODE = $save;
      return $arr;
    }
    
    function _init($parentDriver)
    {
        $parentDriver->hasTransactions = false;
        $parentDriver->hasInsertID = true;
    }

    function _insertid()
    {
        return sqlite_last_insert_rowid($this->_connectionID);
    }
    
    function _affectedrows()
    {
        return sqlite_changes($this->_connectionID);
    }
    
    function ErrorMsg() 
    {
        if ($this->_logsql) {
            return $this->_errorMsg;
        }
        return $this->_errorNo;
    }
 
    function ErrorNo() 
    {
        return $this->_errorNo;
    }
    
    function SQLDate($fmt, $col=false)
    {
        $fmt = $this->qstr($fmt);
        return ($col) ? "adodb_date2($fmt,$col)" : "adodb_date($fmt)";
    }
    
    /**
     * _connect
     *
     * @access private
     * @return bool
     */
    function _connect($argHostname, $argUsername, $argPassword, $argDatabasename)
    {
        if (!class_exists('Text_DB')) {
            return null;
        }

        $this->_connectionID = new Text_DB($argDatabasename);

        if ($this->_connectionID === false) {
            return false;
        }

        return true;
    }
    
    /**
     * _query
     *
     * returns query ID if successful, otherwise false
     *
     * @access private
     * @param string $sql
     */
    function _query($sql, $inputarr = false)
    {
        if ($inputarr !== false) {
            $this->_errorNo = "ERROR: sorry, place folder is not implement";
            return false;
        }

        $result = $this->_connectionID->executeQuery($sql);
        if (!$result) {
            $this->_errorNo = "ERROR: _query {$sql}";
        }
        
        return $result;
    }
    
    function &SelectLimit($sql,$nrows=-1,$offset=-1,$inputarr=false,$secs2cache=0) 
    {
        $offsetStr = ($offset >= 0) ? " OFFSET $offset" : '';
        $limitStr  = ($nrows >= 0)  ? " LIMIT $nrows" : ($offset >= 0 ? ' LIMIT 999999999' : '');
        if ($secs2cache)
            $rs =& $this->CacheExecute($secs2cache,$sql."$limitStr$offsetStr",$inputarr);
        else
            $rs =& $this->Execute($sql."$limitStr$offsetStr",$inputarr);
            
        return $rs;
    }
    
    /*
        This algorithm is not very efficient, but works even if table locking
        is not available.
        
        Will return false if unable to generate an ID after $MAXLOOPS attempts.
    */
    var $_genSeqSQL = "create table %s (id integer)";
    
    function GenID($seq='adodbseq',$start=1)
    {   
        // if you have to modify the parameter below, your database is overloaded,
        // or you need to implement generation of id's yourself!
        $MAXLOOPS = 100;
        //$this->debug=1;
        while (--$MAXLOOPS>=0) {
            @($num = $this->GetOne("select id from $seq"));
            if ($num === false) {
                $this->Execute(sprintf($this->_genSeqSQL ,$seq));   
                $start -= 1;
                $num = '0';
                $ok = $this->Execute("insert into $seq values($start)");
                if (!$ok) return false;
            } 
            $this->Execute("update $seq set id=id+1 where id=$num");
            
            if ($this->affected_rows() > 0) {
                $num += 1;
                $this->genID = $num;
                return $num;
            }
        }
        if ($fn = $this->raiseErrorFn) {
            $fn($this->databaseType,'GENID',-32000,"Unable to generate unique id after $MAXLOOPS attempts",$seq,$num);
        }
        return false;
    }

    function CreateSequence($seqname='adodbseq',$start=1)
    {
        if (empty($this->_genSeqSQL)) return false;
        $ok = $this->Execute(sprintf($this->_genSeqSQL,$seqname));
        if (!$ok) return false;
        $start -= 1;
        return $this->Execute("insert into $seqname values($start)");
    }
    
    var $_dropSeqSQL = 'drop table %s';
    function DropSequence($seqname)
    {
        if (empty($this->_dropSeqSQL)) return false;
        return $this->Execute(sprintf($this->_dropSeqSQL,$seqname));
    }
    
    // returns true or false
    function _close()
    {
        return @sqlite_close($this->_connectionID);
    }

    function &MetaIndexes($table, $primary = FALSE, $owner=false)
    {
        $false = false;
        // save old fetch mode
        global $ADODB_FETCH_MODE;
        $save = $ADODB_FETCH_MODE;
        $ADODB_FETCH_MODE = ADODB_FETCH_NUM;
        if ($this->fetchMode !== FALSE) {
               $savem = $this->SetFetchMode(FALSE);
        }
        $SQL=sprintf("SELECT name,sql FROM sqlite_master WHERE type='index' AND tbl_name='%s'", strtolower($table));
        $rs = $this->Execute($SQL);
        if (!is_object($rs)) {
            if (isset($savem)) 
                $this->SetFetchMode($savem);
            $ADODB_FETCH_MODE = $save;
            return $false;
        }

        $indexes = array ();
        while ($row = $rs->FetchRow()) {
            if ($primary && preg_match("/primary/i",$row[1]) == 0) continue;
            if (!isset($indexes[$row[0]])) {

            $indexes[$row[0]] = array(
                   'unique' => preg_match("/unique/i",$row[1]),
                   'columns' => array());
            }
            /**
              * There must be a more elegant way of doing this,
              * the index elements appear in the SQL statement
              * in cols[1] between parentheses
              * e.g CREATE UNIQUE INDEX ware_0 ON warehouse (org,warehouse)
              */
            $cols = explode("(",$row[1]);
            $cols = explode(")",$cols[1]);
            array_pop($cols);
            $indexes[$row[0]]['columns'] = $cols;
        }
        if (isset($savem)) { 
            $this->SetFetchMode($savem);
            $ADODB_FETCH_MODE = $save;
        }
        return $indexes;
    }

}

/**
 * Class Name: Recordset
 * 
 */
class ADORecordset_textdb extends ADORecordSet {

    var $databaseType = "textdb";
    var $bind = false;

    function ADORecordset_textdb($queryID, $mode = false)
    {
        
        if ($mode === false) { 
            global $ADODB_FETCH_MODE;
            $mode = $ADODB_FETCH_MODE;
        }

        switch($mode) {
            case ADODB_FETCH_NUM:
                $this->fetchMode = SQLITE_NUM;
                break;
            case ADODB_FETCH_ASSOC:
                $this->fetchMode = SQLITE_ASSOC;
                break;
            default: $this->fetchMode = SQLITE_BOTH;
                break;
        }

        $this->adodbFetchMode = $mode;
        
        $this->_queryID = $queryID;
    
        $this->_inited = true;
        $this->fields = array();

        if (is_object($queryID)) {
            $this->_currentRow = 0;
            $this->EOF = !$this->_fetch();
            @$this->_initrs();
        } else {
            $this->_numOfRows = 0;
            $this->_numOfFields = 0;
            $this->EOF = true;
        }
        
        return $this->_queryID;
    }


    function &FetchField($fieldOffset = -1)
    {
        $fld = new ADOFieldObject;
        $columns = $this->_queryID->getColumnNames();
        $fld->name = $columns[$fieldOffset];
        $fld->type = 'VARCHAR';
        $fld->max_length = -1;

        return $fld;
    }
    
    function _initrs()
    {
        $this->_numOfRows = $this->_queryID->getRowCount();
        $this->_numOfFields = $this->_queryID->getRowSize();
    }

    function Fields($colname)
    {
        if ($this->fetchMode != SQLITE_NUM) {
            return $this->fields[$colname];
        }

        if (!$this->bind) {

            $this->bind = array();

            for ($i=0; $i < $this->_numOfFields; $i++) {
                $o = $this->FetchField($i);
                $this->bind[strtoupper($o->name)] = $i;
            }

        }
        
        return $this->fields[$this->bind[strtoupper($colname)]];
   }
    
   function _seek($row)
   {
        return sqlite_seek($this->_queryID, $row);
   }

    function _fetch($ignore_fields = false) 
    {
        $result = $this->_queryID->next();
        $this->fields = $this->_queryID->getCurrentValuesAsHash();

        return $result;
    }
    
    function _close() 
    {
    }

}
?>
